from typing import Any, List

from pydantic.validators import int_validator, str_validator

from .code_exception import CodeException


class VKAPIError(CodeException):
    def __init__(self, *, error_msg: Any, request_params: List[dict]):
        super().__init__(error_msg)
        self.description = str_validator(error_msg)
        self.params = {item["key"]: item["value"] for item in request_params}


class CaptchaError(VKAPIError[14]):  # type: ignore
    def __init__(self, *, captcha_sid: Any, captcha_img: Any, **kwargs):
        super().__init__(**kwargs)
        self.sid = int_validator(captcha_sid)
        self.img = str_validator(captcha_img)



import logging

from choicelib import choice_in_order

json = choice_in_order(["ujson", "hyperjson", "orjson"], do_import=True, default="json")
logger = choice_in_order(["loguru"], do_import=True, default="logging")

if logger.__name__ == "logging":
    logger = logging.getLogger("vkbottle")
elif logger.__name__ == "loguru":
    logger = getattr(logger, "logger")



from io import StringIO
from typing import TYPE_CHECKING, Any, List, Optional, Union

from vkbottle_types import StatePeer
from vkbottle_types.objects import MessagesMessage, UsersUser

if TYPE_CHECKING:

    from vkbottle.api import ABCAPI, API


class BaseMessageMin(MessagesMessage):
    unprepared_ctx_api: Optional[Any] = None
    state_peer: Optional["StatePeer"] = None

    @property
    def ctx_api(self) -> Union["ABCAPI", "API"]:
        return getattr(self, "unprepared_ctx_api")

    async def get_user(self, raw_mode: bool = False, **kwargs) -> Union["UsersUser", dict]:
        raw_user = (await self.ctx_api.request("users.get", {"user_ids": self.from_id, **kwargs}))[
            "response"
        ][0]
        return raw_user if raw_mode else UsersUser(**raw_user)

    async def answer(
        self,
        message: Optional[str] = None,
        attachment: Optional[str] = None,
        user_id: Optional[int] = None,
        random_id: Optional[int] = 0,
        peer_id: Optional[int] = None,
        peer_ids: Optional[List[int]] = None,
        domain: Optional[str] = None,
        chat_id: Optional[int] = None,
        user_ids: Optional[List[int]] = None,
        lat: Optional[float] = None,
        long: Optional[float] = None,
        reply_to: Optional[int] = None,
        forward_messages: Optional[List[int]] = None,
        forward: Optional[str] = None,
        sticker_id: Optional[int] = None,
        group_id: Optional[int] = None,
        keyboard: Optional[str] = None,
        template: Optional[str] = None,
        payload: Optional[str] = None,
        content_source: Optional[str] = None,
        dont_parse_links: Optional[bool] = None,
        disable_mentions: Optional[bool] = None,
        intent: Optional[str] = None,
        subscribe_id: Optional[int] = None,
        **kwargs,
    ) -> int:
        locals().update(kwargs)

        data = {k: v for k, v in locals().items() if k not in ("self", "kwargs") and v is not None}
        required_params = ("peer_id", "user_id", "domain", "chat_id", "user_ids")
        if not any(data.get(param) for param in required_params):
            data["peer_id"] = self.peer_id

        stream = StringIO(message)
        while True:
            msg = stream.read(4096)
            if msg:
                data["message"] = msg
            response = (await self.ctx_api.request("messages.send", data))["response"]
            if stream.tell() == len(message or ""):
                break
        return response


BaseMessageMin.update_forward_refs()



from typing import TYPE_CHECKING, List, Optional

from ..base import BaseMessageMin

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI


class MessageMin(BaseMessageMin):
    user_id: Optional[int] = None

    async def reply(
        self,
        message: Optional[str] = None,
        attachment: Optional[str] = None,
        **kwargs,
    ) -> int:
        locals().update(kwargs)

        data = {k: v for k, v in locals().items() if k not in ("self", "kwargs") and v is not None}
        data["peer_id"] = self.peer_id
        data["reply_to"] = self.id

        return await self.answer(**data)

    async def forward(
        self,
        message: Optional[str] = None,
        attachment: Optional[str] = None,
        forward_message_ids: Optional[List[int]] = None,
        **kwargs,
    ) -> int:
        locals().update(kwargs)

        data = {
            k: v
            for k, v in locals().items()
            if k not in ("self", "kwargs", "forward_message_ids") and v is not None
        }
        if not forward_message_ids:
            forward_message_ids = [self.id]

        data["forward_messages"] = forward_message_ids

        return await self.answer(**data)


MessageMin.update_forward_refs()


async def message_min(message_id: int, ctx_api: "ABCAPI") -> "MessageMin":
    message_object = (await ctx_api.request("messages.getById", {"message_ids": message_id}))[
        "response"
    ]["items"][0]
    message = MessageMin(**message_object)
    setattr(message, "unprepared_ctx_api", ctx_api)
    return message



from io import StringIO
from typing import TYPE_CHECKING, Any, List, Optional, Union

from vkbottle_types.events.objects.group_event_objects import MessageEventObject

from vkbottle import StatePeer
from vkbottle.tools.dev.event_data import OpenAppEvent, OpenLinkEvent, ShowSnackbarEvent

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI, API

    EventDataType = Union[ShowSnackbarEvent, OpenAppEvent, OpenLinkEvent]


class MessageEventMin(MessageEventObject):
    unprepared_ctx_api: Optional[Any] = None
    state_peer: Optional["StatePeer"] = None
    group_id: Optional[int] = None

    def __init__(self, **event):
        data = event["object"]
        data["group_id"] = event["group_id"]

        super().__init__(**data)

    @property
    def ctx_api(self) -> Union["ABCAPI", "API"]:
        return getattr(self, "unprepared_ctx_api")

    async def send_message_event_answer(self, event_data: "EventDataType", **kwargs) -> int:
        data = dict(
            event_id=self.event_id,
            user_id=self.user_id,
            peer_id=self.peer_id,
            event_data=event_data.json(),
        )
        data.update(kwargs)
        return (await self.ctx_api.request("messages.sendMessageEventAnswer", data))["response"]

    async def show_snackbar(self, text: str) -> int:
        return await self.send_message_event_answer(ShowSnackbarEvent(text=text))

    async def open_link(self, link: str) -> int:
        return await self.send_message_event_answer(OpenLinkEvent(link=link))

    async def open_app(self, app_id: int, app_hash: str, owner_id: Optional[int] = None) -> int:
        return await self.send_message_event_answer(
            OpenAppEvent(app_id=app_id, hash=app_hash, owner_id=owner_id)
        )

    async def edit_message(
        self,
        message: Optional[str] = None,
        lat: Optional[float] = None,
        long: Optional[float] = None,
        attachment: Optional[str] = None,
        keep_forward_messages: Optional[bool] = None,
        keep_snippets: Optional[bool] = None,
        dont_parse_links: Optional[bool] = None,
        template: Optional[str] = None,
        keyboard: Optional[str] = None,
        **kwargs,
    ) -> int:
        locals().update(kwargs)

        data = {k: v for k, v in locals().items() if k not in ("self", "kwargs") and v is not None}
        data["peer_id"] = self.peer_id
        data["conversation_message_id"] = self.conversation_message_id
        return (await self.ctx_api.request("messages.edit", data))["response"]

    async def send_message(
        self,
        message: Optional[str] = None,
        attachment: Optional[str] = None,
        user_id: Optional[int] = None,
        random_id: Optional[int] = 0,
        peer_id: Optional[int] = None,
        peer_ids: Optional[List[int]] = None,
        domain: Optional[str] = None,
        chat_id: Optional[int] = None,
        user_ids: Optional[List[int]] = None,
        lat: Optional[float] = None,
        long: Optional[float] = None,
        reply_to: Optional[int] = None,
        forward_messages: Optional[List[int]] = None,
        forward: Optional[str] = None,
        sticker_id: Optional[int] = None,
        group_id: Optional[int] = None,
        keyboard: Optional[str] = None,
        template: Optional[str] = None,
        payload: Optional[str] = None,
        content_source: Optional[str] = None,
        dont_parse_links: Optional[bool] = None,
        disable_mentions: Optional[bool] = None,
        intent: Optional[str] = None,
        subscribe_id: Optional[int] = None,
        **kwargs,
    ) -> int:
        locals().update(kwargs)

        data = {k: v for k, v in locals().items() if k not in ("self", "kwargs") and v is not None}
        required_params = ("peer_id", "user_id", "domain", "chat_id", "user_ids")
        if not any(data.get(param) for param in required_params):
            data["peer_id"] = self.peer_id

        stream = StringIO(message)
        while True:
            msg = stream.read(4096)
            if msg:
                data["message"] = msg
            response = (await self.ctx_api.request("messages.send", data))["response"]
            if stream.tell() == len(message or ""):
                break
        return response


MessageEventMin.update_forward_refs()



from typing import TYPE_CHECKING, Optional, Union

from vkbottle.dispatch import Router
from vkbottle.framework.abc_blueprint import ABCBlueprint
from vkbottle.framework.bot.bot import Bot
from vkbottle.framework.labeler import BotLabeler
from vkbottle.modules import logger

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI, API
    from vkbottle.dispatch import ABCStateDispenser
    from vkbottle.polling import ABCPolling


class BotBlueprint(ABCBlueprint):
    def __init__(
        self,
        name: Optional[str] = None,
        labeler: Optional[BotLabeler] = None,
        router: Optional[Router] = None,
    ):
        if name is not None:
            self.name = name

        self.labeler = labeler or BotLabeler()
        self.router: Router = router or Router()
        self.constructed = False

    def construct(
        self,
        api: Union["ABCAPI", "API"],
        polling: "ABCPolling",
        state_dispenser: "ABCStateDispenser",
    ) -> "BotBlueprint":
        self.api = api
        self.polling = polling
        self.state_dispenser = state_dispenser
        self.constructed = True
        return self

    def load(self, framework: "Bot") -> "BotBlueprint":
        framework.labeler.load(self.labeler)  # type: ignore
        logger.debug(f"Blueprint {self.name!r} loaded")
        return self.construct(framework.api, framework.polling, framework.state_dispenser)

    @property
    def on(self) -> BotLabeler:
        return self.labeler



from abc import ABC
from typing import TYPE_CHECKING, Union

from .base import BaseUploader

if TYPE_CHECKING:
    from .base import Bytes


class DocUploader(BaseUploader, ABC):
    NAME = "doc.txt"

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("docs.getUploadServer", kwargs))["response"]

    async def upload(
        self, title: str, file_source: Union[str, "Bytes"], **params
    ) -> Union[str, dict]:
        server = await self.get_server(**params)
        data = await self.read(file_source)
        file = self.get_bytes_io(data, name=title)

        uploader = await self.upload_files(server["upload_url"], {"file": file})

        doc = (
            await self.api.request(
                "docs.save",
                {"title": title, **uploader, **params},
            )
        )["response"]
        doc_type = doc["type"]

        if self.generate_attachment_strings:
            return self.generate_attachment_string(
                doc_type,
                doc[doc_type]["owner_id"],
                doc[doc_type]["id"],
                doc[doc_type].get("access_key"),
            )
        return doc

    @property
    def attachment_name(self) -> str:
        return self.with_name or self.NAME


class DocWallUploader(DocUploader):
    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("docs.getWallUploadServer", kwargs))["response"]


class DocMessagesUploader(DocUploader):
    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("docs.getMessagesUploadServer", kwargs))["response"]


class VoiceMessageUploader(DocUploader):
    async def get_server(self, **kwargs) -> dict:
        return (
            await self.api.request(
                "docs.getMessagesUploadServer", {"type": "audio_message", **kwargs}
            )
        )["response"]


class GraffitiUploader(DocUploader):
    async def get_server(self, **kwargs) -> dict:
        return (
            await self.api.request("docs.getMessagesUploadServer", {"type": "graffiti", **kwargs})
        )["response"]


__all__ = (
    "DocUploader",
    "DocWallUploader",
    "DocMessagesUploader",
    "VoiceMessageUploader",
    "GraffitiUploader",
)



from .abc import ABCResponseValidator
from .json_validator import JSONResponseValidator
from .vk_api_error_validator import VKAPIErrorResponseValidator

DEFAULT_RESPONSE_VALIDATORS = [
    JSONResponseValidator(),
    VKAPIErrorResponseValidator(),
]

__all__ = (
    "ABCResponseValidator",
    "DEFAULT_RESPONSE_VALIDATORS",
    "JSONResponseValidator",
    "VKAPIErrorResponseValidator",
)



from .action import ABCAction, Callback, Location, OpenLink, Text, VKApps, VKPay
from .color import KeyboardButtonColor
from .keyboard import Keyboard

EMPTY_KEYBOARD = Keyboard().get_json()

__all__ = (
    "ABCAction",
    "Callback",
    "EMPTY_KEYBOARD",
    "Keyboard",
    "KeyboardButtonColor",
    "Location",
    "OpenLink",
    "Text",
    "VKApps",
    "VKPay",
)



import asyncio
from typing import TYPE_CHECKING, AsyncIterator, Optional

from vkbottle.exception_factory import ErrorHandler
from vkbottle.modules import logger

from .abc import ABCPolling

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI
    from vkbottle.exception_factory import ABCErrorHandler


class UserPolling(ABCPolling):
    """User Polling class
    Documentation: https://github.com/vkbottle/vkbottle/blob/master/docs/low-level/polling/polling.md
    """

    def __init__(
        self,
        api: Optional["ABCAPI"] = None,
        user_id: Optional[int] = None,
        wait: Optional[int] = None,
        mode: Optional[int] = None,
        rps_delay: Optional[int] = None,
        error_handler: Optional["ABCErrorHandler"] = None,
    ):
        self._api = api
        self.error_handler = error_handler or ErrorHandler()
        self.user_id = user_id
        self.wait = wait or 15
        self.mode = mode or 234
        self.rps_delay = rps_delay or 0
        self.stop = False

    async def get_event(self, server: dict) -> dict:
        logger.debug("Making long request to get event with longpoll...")
        return await self.api.http_client.request_json(
            "https://{}?act=a_check&key={}&ts={}&wait={}&mode={}&rps_delay={}".format(
                server["server"],
                server["key"],
                server["ts"],
                self.wait,
                self.mode,
                self.rps_delay,
            ),
            method="POST",
        )

    async def get_server(self) -> dict:
        logger.debug("Getting polling server...")
        if self.user_id is None:
            self.user_id = (await self.api.request("users.get", {}))["response"][0]["id"]
        return (await self.api.request("messages.getLongPollServer", {}))["response"]

    async def listen(self) -> AsyncIterator[dict]:  # type: ignore
        logger.debug("Starting listening to longpoll")
        server: Optional[dict] = None
        while not self.stop:
            try:
                if not server:
                    server = await self.get_server()
                event = await self.get_event(server)
                if not event.get("ts"):
                    server = await self.get_server()
                    continue
                server["ts"] = event["ts"]
                yield event
            except (TimeoutError, asyncio.exceptions.TimeoutError):
                logger.error(
                    f"Looks like VK are dead, sleeping {self.wait} seconds and trying again..."
                )
                await asyncio.sleep(self.wait)
                server = None
            except BaseException as e:
                await self.error_handler.handle(e)

    def construct(
        self, api: "ABCAPI", error_handler: Optional["ABCErrorHandler"] = None
    ) -> "UserPolling":
        self._api = api
        if error_handler is not None:
            self.error_handler = error_handler
        return self

    @property
    def api(self) -> "ABCAPI":
        if self._api is None:
            raise NotImplementedError(
                "You must construct polling with API before try to access api property of Polling"
            )
        return self._api

    @api.setter
    def api(self, new_api: "ABCAPI"):
        self._api = new_api



from typing import List
from warnings import warn

from vkbottle.modules import json


def keyboard_gen(pattern: List[List[dict]], one_time: bool = False, inline: bool = False):
    """Simple keyboard generator
    :param pattern: Keyboard simple pattern, check github readme
    :param one_time: Should keyboard be hidden after first use?
    :param inline: Should keyboard be inline?
    :return: VK Api Keyboard JSON
    """
    warn("keyboard_gen generator is obsolete, use Keyboard instead", DeprecationWarning)

    rows = pattern
    buttons = []

    for row in rows:
        row_buttons = []
        for button in row:
            action = {}
            fields = {}

            action["type"] = button.get("type", "text")
            action.update({k: v for k, v in button.items() if k not in ("type", "text", "color")})

            if action["type"] == "text":
                action["label"] = button.get("text", button.get("label"))

            if button.get("color"):
                fields["color"] = button["color"]

            row_button = {"action": action, **fields}

            row_buttons.append(row_button)
        buttons.append(row_buttons)

    return str(
        json.dumps(dict(one_time=one_time, inline=inline, buttons=buttons))
        .encode("utf-8")
        .decode("utf-8")
    )



import os
import sys

from watchgod import awatch

from vkbottle.modules import logger

_startup_cwd = os.getcwd()


def restart():
    """https://github.com/cherrypy/cherrypy/blob/0857fa81eb0ab647c7b59a019338bab057f7748b/cherrypy/process/wspbus.py#L305"""
    args = sys.argv[:]
    logger.debug("Restarting: %s" % " ".join(args))
    args.insert(0, sys.executable)
    if sys.platform == "win32":
        args = ['"%s"' % arg for arg in args]

    os.chdir(_startup_cwd)
    os.execv(sys.executable, args)


async def watch_to_reload(check_dir: str):
    """
    Coro which see changes in your code and restart him.
    :return:
    """
    async for _ in awatch(check_dir):
        logger.info("Changes were found. Restarting...")
        restart()



from .abc import ABCRequestValidator
from .translate_friendly_types_validator import TranslateFriendlyTypesRequestValidator

DEFAULT_REQUEST_VALIDATORS = [TranslateFriendlyTypesRequestValidator()]

__all__ = (
    "ABCRequestValidator",
    "DEFAULT_REQUEST_VALIDATORS",
    "TranslateFriendlyTypesRequestValidator",
)



from typing import TYPE_CHECKING, Any, Dict, NoReturn, Type, Union

from vkbottle.exception_factory import CaptchaError, VKAPIError

from .abc import ABCResponseValidator

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI, API


SPECIFIC_ERRORS: Dict[int, Type[VKAPIError]] = {14: CaptchaError}


class VKAPIErrorResponseValidator(ABCResponseValidator):
    """Default vk api error response validator
    Documentation: https://github.com/vkbottle/vkbottle/blob/master/docs/low-level/api/response-validator.md
    """

    async def validate(
        self,
        method: str,
        data: dict,
        response: Any,
        ctx_api: Union["ABCAPI", "API"],
    ) -> Union[Any, NoReturn]:
        if "error" not in response:
            return response

        if ctx_api.ignore_errors:
            return None

        error = response["error"]
        code = error.pop("error_code")

        exception = SPECIFIC_ERRORS.get(code, VKAPIError[code])
        raise exception(**error)



from typing import List, Optional, Union

from vkbottle.modules import json


class TemplateElement:
    """Easy template element generator"""

    def __init__(
        self,
        title: Optional[str] = None,
        description: Optional[str] = None,
        photo_id: Optional[str] = None,
        buttons: Optional[Union[List[dict], str]] = None,
        action: Optional[dict] = None,
    ):
        assert buttons, "Buttons are required"
        assert photo_id or title, "photo_id or title is required"

        if isinstance(buttons, str):
            buttons = json.loads(buttons)

        elif isinstance(buttons, dict):
            buttons = buttons.get("buttons")[0]  # taking only first row

        self.raw: dict = {k: v for k, v in locals().items() if v is not None and k != "self"}



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Union

if TYPE_CHECKING:
    from vkbottle_types.events import Event


class ABCHandler(ABC):
    blocking: bool

    @abstractmethod
    async def filter(self, event: "Event") -> Union[dict, bool]:
        pass

    @abstractmethod
    async def handle(self, event: "Event", **context) -> Any:
        pass



from .base_exceptions import CaptchaError, VKAPIError
from .code_exception import CodeException
from .error_handler import ABCErrorHandler, ErrorHandler
from .swear_handler import swear

__all__ = (
    "ABCErrorHandler",
    "CaptchaError",
    "CodeException",
    "ErrorHandler",
    "swear",
    "VKAPIError",
)



from .abc import ABCRouter
from .base import Router
from .dispenser import ABCStateDispenser, BaseStateGroup, BuiltinStateDispenser, StatePeer
from .handlers import ABCHandler
from .middlewares import BaseMiddleware, MiddlewareError
from .return_manager import BaseReturnManager
from .rules import ABCRule, AndRule, NotRule, OrRule
from .views import ABCDispenseView, ABCView

__all__ = (
    "ABCDispenseView",
    "ABCHandler",
    "ABCRouter",
    "ABCRule",
    "ABCStateDispenser",
    "ABCView",
    "AndRule",
    "BaseMiddleware",
    "BaseReturnManager",
    "BaseStateGroup",
    "BuiltinStateDispenser",
    "MiddlewareError",
    "NotRule",
    "OrRule",
    "Router",
    "StatePeer",
)



from typing import TYPE_CHECKING, List, Optional, Union

from .base import BaseUploader

if TYPE_CHECKING:
    from .base import Bytes


class VideoUploader(BaseUploader):
    NAME = "video.mp4"

    @property
    def attachment_name(self) -> str:
        return self.with_name or self.NAME

    async def upload(
        self,
        file_source: Optional[Union[str, "Bytes"]] = None,
        group_id: Optional[int] = None,
        **params,
    ) -> Union[str, List[dict], dict]:
        server = await self.get_server(group_id=group_id)
        assert (
            file_source is not None or "link" in params
        ), "file_source or link to video must be set"

        if "link" in params and not file_source:
            return await self.api.http_client.request_json(
                server["upload_url"], method="GET", params=params
            )

        data = await self.read(file_source)  # type: ignore
        file = self.get_bytes_io(data)
        video = await self.upload_files(server["upload_url"], {"video_file": file})

        if self.generate_attachment_strings:
            return self.generate_attachment_string("video", video["owner_id"], video["video_id"])
        return video

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("video.save", kwargs))["response"]


__all__ = ("VideoUploader",)



from typing import Dict, Tuple, Type, TypeVar, Union, cast, no_type_check, overload

T = TypeVar("T", bound=Type["CodeException"])


class CodeException(Exception):
    code: int
    __code_specified__: bool = False
    __exceptions__: Dict[int, Type["CodeException"]] = {}

    @no_type_check
    def __new__(cls, *args, **kwargs):
        if not cls.__code_specified__:
            raise TypeError("exception code is not specified")

        return super().__new__(cls, *args, **kwargs)

    @no_type_check
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.__exceptions__ = {}

    @overload
    def __class_getitem__(cls: T, code: int) -> T:
        ...

    @overload
    def __class_getitem__(cls: T, code: Tuple[int, ...]) -> Tuple[T, ...]:
        ...

    def __class_getitem__(cls: T, code: Union[int, Tuple[int, ...]]) -> Union[T, Tuple[T, ...]]:
        if cls.__code_specified__:
            raise TypeError("exception code already specified")

        if isinstance(code, tuple):
            return tuple(cls._get_exception(c) for c in code)

        return cls._get_exception(code)

    @classmethod
    def _get_exception(cls: T, code: int) -> T:
        if code in cls.__exceptions__:
            return cast(T, cls.__exceptions__[code])

        return cls._register_exception(code)

    @classmethod
    def _register_exception(cls: T, code: int) -> T:
        name = f"{cls.__name__}_{code}"
        exception = cast(T, type(name, (cls,), {}))

        exception.code = code
        exception.__code_specified__ = True
        cls.__exceptions__[code] = exception

        return exception



from .abc import ABCFramework
from .abc_blueprint import ABCBlueprint
from .bot import Bot, BotBlueprint, run_multibot
from .user import User, UserBlueprint

__all__ = (
    "ABCBlueprint",
    "ABCFramework",
    "Bot",
    "BotBlueprint",
    "run_multibot",
    "User",
    "UserBlueprint",
)



from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Iterable,
    List,
    NamedTuple,
    NoReturn,
    Optional,
    Union,
)

import vkbottle_types
from vkbottle_types.categories import APICategories

from vkbottle.http import SingleAiohttpClient
from vkbottle.modules import logger

from .abc import ABCAPI
from .request_rescheduler import BlockingRequestRescheduler
from .request_validator import DEFAULT_REQUEST_VALIDATORS
from .response_validator import DEFAULT_RESPONSE_VALIDATORS
from .token_generator import get_token_generator

if TYPE_CHECKING:
    from vkbottle.http import ABCHTTPClient

    from .request_rescheduler import ABCRequestRescheduler
    from .request_validator import ABCRequestValidator
    from .response_validator import ABCResponseValidator
    from .token_generator import Token

APIRequest = NamedTuple("APIRequest", [("method", str), ("data", dict)])


class API(ABCAPI, APICategories):
    """Default API instance
    Documentation: https://github.com/vkbottle/vkbottle/blob/master/docs/low-level/api/api.md
    """

    API_URL = vkbottle_types.API_URL
    API_VERSION = vkbottle_types.API_VERSION
    APIRequest = APIRequest

    def __init__(
        self,
        token: "Token",
        ignore_errors: bool = False,
        http_client: Optional["ABCHTTPClient"] = None,
        request_rescheduler: Optional["ABCRequestRescheduler"] = None,
    ):
        self.token_generator = get_token_generator(token)
        self.ignore_errors = ignore_errors
        self.http_client = http_client or SingleAiohttpClient()
        self.request_rescheduler = request_rescheduler or BlockingRequestRescheduler()
        self.response_validators: List["ABCResponseValidator"] = DEFAULT_RESPONSE_VALIDATORS
        self.request_validators: List["ABCRequestValidator"] = DEFAULT_REQUEST_VALIDATORS  # type: ignore

    async def request(self, method: str, data: dict) -> dict:
        """Makes a single request opening a session"""
        data = await self.validate_request(data)

        async with self.token_generator as token:
            response = await self.http_client.request_text(
                self.API_URL + method,
                method="POST",
                data=data,  # type: ignore
                params={"access_token": token, "v": self.API_VERSION},
            )
        logger.debug("Request {} with {} data returned {}".format(method, data, response))
        return await self.validate_response(method, data, response)

    async def request_many(
        self, requests: Iterable[APIRequest]  # type: ignore
    ) -> AsyncIterator[dict]:
        """Makes many requests opening one session"""
        for request in requests:
            method, data = request.method, await self.validate_request(request.data)  # type: ignore
            async with self.token_generator as token:
                response = await self.http_client.request_json(
                    self.API_URL + method,
                    method="POST",
                    data=data,  # noqa
                    params={"access_token": token, "v": self.API_VERSION},  # noqa
                )
            logger.debug("Request {} with {} data returned {}".format(method, data, response))
            yield await self.validate_response(method, data, response)

    async def validate_response(
        self, method: str, data: dict, response: Union[dict, str]
    ) -> Union[Any, NoReturn]:
        """Validates response from VK,
        to change validations change API.response_validators (list of ResponseValidator's)"""
        for validator in self.response_validators:
            response = await validator.validate(method, data, response, self)
        logger.debug("API response was validated")
        return response  # type: ignore

    async def validate_request(self, request: dict) -> dict:
        """Validates requests from VK,
        to change validations change API.request_validators (list of RequestValidator's)"""
        for validator in self.request_validators:
            request = await validator.validate(request)
        logger.debug("API request was validated")
        return request  # type: ignore

    @property
    def api_instance(self) -> "API":
        return self

    def __repr__(self) -> str:
        return f"<API token_generator={self.token_generator}...>"



from abc import ABC
from typing import TYPE_CHECKING, List, Union

from .base import BaseUploader

if TYPE_CHECKING:
    from .base import Bytes


class PhotoUploader(BaseUploader, ABC):
    NAME = "picture.jpg"

    @property
    def attachment_name(self) -> str:
        return self.with_name or self.NAME


class PhotoToAlbumUploader(PhotoUploader):
    async def upload(
        self, album_id: int, paths_like: Union[List[Union[str, "Bytes"]], str, "Bytes"], **params
    ) -> Union[str, List[Union[str, dict]]]:
        if not isinstance(paths_like, list):
            paths_like = [paths_like]

        server = await self.get_server(album_id=album_id, **params)
        files = {}

        for i, file_source in enumerate(paths_like):
            data = await self.read(file_source)
            files[f"file{i+1}"] = self.get_bytes_io(data)

        uploader = await self.upload_files(server["upload_url"], files)
        photos = (
            await self.api.request("photos.save", {"album_id": album_id, **uploader, **params})
        )["response"]

        if self.generate_attachment_strings:
            return [
                self.generate_attachment_string(
                    "photo", photo["owner_id"], photo["id"], photo.get("access_key")
                )
                for photo in photos
            ]
        return photos

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("photos.getUploadServer", kwargs))["response"]


class PhotoWallUploader(PhotoUploader):
    async def upload(self, file_source: Union[str, "Bytes"], **params) -> Union[str, List[dict]]:
        server = await self.get_server(**params)
        data = await self.read(file_source)
        file = self.get_bytes_io(data)

        uploader = await self.upload_files(server["upload_url"], {"photo": file})
        photos = (await self.api.request("photos.saveWallPhoto", {**uploader, **params}))[
            "response"
        ]

        if self.generate_attachment_strings:
            return self.generate_attachment_string(
                "photo", photos[0]["owner_id"], photos[0]["id"], photos[0].get("access_key")
            )
        return photos

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("photos.getWallUploadServer", {}))["response"]


class PhotoFaviconUploader(PhotoUploader):
    async def upload(self, file_source: Union[str, "Bytes"], **params) -> Union[str, dict]:
        owner_id = await self.get_owner_id(params)
        server = await self.get_server(owner_id=owner_id)
        data = await self.read(file_source)
        file = self.get_bytes_io(data)

        uploader = await self.upload_files(server["upload_url"], {"photo": file})
        photo = (await self.api.request("photos.saveOwnerPhoto", {**uploader, **params}))[
            "response"
        ]

        if self.generate_attachment_strings:
            return self.generate_attachment_string(
                "wall", owner_id, photo["post_id"], photo.get("access_key")
            )
        return photo

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("photos.getOwnerPhotoUploadServer", kwargs))["response"]


class PhotoMessageUploader(PhotoUploader):
    async def upload(self, file_source: Union[str, "Bytes"], **params) -> Union[str, List[dict]]:
        server = await self.get_server(**params)
        data = await self.read(file_source)
        file = self.get_bytes_io(data)

        uploader = await self.upload_files(server["upload_url"], {"photo": file})
        photo = (await self.api.request("photos.saveMessagesPhoto", {**uploader, **params}))[
            "response"
        ]

        if self.generate_attachment_strings:
            return self.generate_attachment_string(
                "photo", photo[0]["owner_id"], photo[0]["id"], photo.get("access_key")
            )
        return photo

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("photos.getMessagesUploadServer", kwargs))["response"]


class PhotoChatFaviconUploader(PhotoUploader):
    async def upload(self, chat_id: int, file_source: Union[str, "Bytes"], **params) -> str:
        server = await self.get_server(chat_id=chat_id, **params)
        data = await self.read(file_source)
        file = self.get_bytes_io(data)

        uploader = await self.upload_files(server["upload_url"], {"photo": file})
        return uploader["response"]

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("photos.getChatUploadServer", kwargs))["response"]


class PhotoMarketUploader(PhotoUploader):
    async def upload(self, file_source: Union[str, "Bytes"], **params) -> dict:
        server = await self.get_server(**params)
        data = await self.read(file_source)
        file = self.get_bytes_io(data)

        uploader = await self.upload_files(server["upload_url"], {"file": file})
        return (await self.api.request("photos.saveMarketPhoto", {**uploader, **params}))[
            "response"
        ]

    async def get_server(self, **kwargs) -> dict:
        return (await self.api.request("photos.getMarketUploadServer", kwargs))["response"]


__all__ = (
    "PhotoUploader",
    "PhotoToAlbumUploader",
    "PhotoWallUploader",
    "PhotoFaviconUploader",
    "PhotoMessageUploader",
    "PhotoChatFaviconUploader",
    "PhotoMarketUploader",
)



from vkbottle.framework.labeler import BotLabeler

from .dispatch.rules import base as rules
from .dispatch.views import bot as views
from .framework.bot import Bot, BotBlueprint, run_multibot
from .tools.dev.mini_types.bot import MessageEventMin, MessageMin

Message = MessageMin
MessageEvent = MessageEventMin
Blueprint = BotBlueprint

__all__ = (
    "Bot",
    "Blueprint",
    "BotLabeler",
    "Message",
    "MessageEvent",
    "run_multibot",
    "rules",
    "views",
)



from typing import Optional

from vkbottle.http.aiohttp import SingleAiohttpClient

MOBILE_APP_ID = 2274003
MOBILE_APP_SECRET = "hHbZxrka2uZ6jB1inYsH"


class AuthError(Exception):
    def __init__(self, typ: str, description: str):
        super().__init__(description)
        self.type = typ
        self.description = description


class UserAuth:
    def __init__(self, client_id: Optional[int] = None, client_secret: Optional[str] = None):
        if client_id is not None and client_secret is not None:
            self.client_id = client_id
            self.client_secret = client_secret
        else:
            self.client_id = MOBILE_APP_ID
            self.client_secret = MOBILE_APP_SECRET

    def _build_oauth_url(self, login: str, password: str) -> str:
        return (
            f"https://oauth.vk.com/token?grant_type=password&client_id={self.client_id}"
            f"&client_secret={self.client_secret}&username={login}&password={password}"
        )

    async def get_token(self, login: str, password: str) -> str:
        url = self._build_oauth_url(login, password)

        client = SingleAiohttpClient()
        response = await client.request_json(url)

        if "access_token" in response:
            return response["access_token"]

        raise AuthError(response["error"], response["error_description"])



from abc import ABC
from typing import Any, Callable, Coroutine, Dict, NamedTuple, Optional, Tuple, Union

HandlerProperty = NamedTuple(
    "HandlerProperty", (("types", Union[type, Tuple[type, ...]]), ("handler", Callable))
)


class BaseReturnManager(ABC):
    def get_handler(self, value: Any) -> Optional[Callable]:
        for types, handler in self.handlers.items():
            if isinstance(value, types):
                return handler

    @property
    def handlers(self) -> Dict[Union[type, Tuple[type, ...]], Callable[[Any], Coroutine]]:
        return {
            v.types: v.handler
            for k, v in vars(self.__class__).items()
            if isinstance(v, HandlerProperty)
        }

    @classmethod
    def instance_of(
        cls, types: Union[type, Tuple[type, ...]]
    ) -> Callable[[Callable], HandlerProperty]:
        def decorator(func: Callable):
            return HandlerProperty(types, func)

        return decorator

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, NoReturn, Union

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI, API


class ABCResponseValidator(ABC):
    """Abstract Response Validator class
    Documentation: https://github.com/vkbottle/vkbottle/blob/master/docs/low-level/api/response-validator.md
    """

    @abstractmethod
    async def validate(
        self,
        method: str,
        data: dict,
        response: Any,
        ctx_api: Union["ABCAPI", "API"],
    ) -> Union[Any, NoReturn]:
        pass



from abc import ABC
from typing import TYPE_CHECKING, Generic, Optional, TypeVar

from vkbottle_types.events import GroupEventType

from vkbottle.dispatch.return_manager.bot import BotMessageReturnHandler
from vkbottle.dispatch.views.abc import ABCMessageView
from vkbottle.tools.dev.mini_types.bot import message_min

if TYPE_CHECKING:
    from vkbottle.tools.dev.mini_types.bot import MessageMin


T_contra = TypeVar("T_contra", contravariant=True)


class ABCBotMessageView(ABCMessageView[dict, T_contra], ABC, Generic[T_contra]):
    def __init__(self):
        super().__init__()
        self.handler_return_manager = BotMessageReturnHandler()

    @staticmethod
    def get_event_type(event: dict) -> str:
        return event["type"]

    @staticmethod
    async def get_message(event, ctx_api) -> "MessageMin":
        return message_min(event, ctx_api)

    async def process_event(self, event: dict) -> bool:
        return GroupEventType(self.get_event_type(event)) == GroupEventType.MESSAGE_NEW


class BotMessageView(ABCBotMessageView["MessageMin"]):
    def get_state_key(self, message: "MessageMin") -> Optional[int]:
        return getattr(message, self.state_source_key, None)



from vkbottle_types import GroupTypes
from vkbottle_types.events import GroupEventType, UserEventType

from .api import (
    ABCAPI,
    API,
    DEFAULT_REQUEST_VALIDATORS,
    DEFAULT_RESPONSE_VALIDATORS,
    ABCRequestRescheduler,
    ABCRequestValidator,
    ABCResponseValidator,
    ABCTokenGenerator,
    BlockingRequestRescheduler,
    ConsistentTokenGenerator,
    SingleTokenGenerator,
    Token,
    get_token_generator,
)
from .dispatch import (
    ABCDispenseView,
    ABCHandler,
    ABCRouter,
    ABCRule,
    ABCStateDispenser,
    ABCView,
    AndRule,
    BaseMiddleware,
    BaseReturnManager,
    BaseStateGroup,
    BuiltinStateDispenser,
    MiddlewareError,
    NotRule,
    OrRule,
    Router,
    StatePeer,
)
from .exception_factory import (
    ABCErrorHandler,
    CaptchaError,
    CodeException,
    ErrorHandler,
    VKAPIError,
    swear,
)
from .framework import (
    ABCBlueprint,
    ABCFramework,
    Bot,
    BotBlueprint,
    User,
    UserBlueprint,
    run_multibot,
)
from .http import ABCHTTPClient, AiohttpClient, SingleAiohttpClient
from .polling import ABCPolling, BotPolling, UserPolling
from .tools import (
    EMPTY_KEYBOARD,
    ABCAction,
    ABCStorage,
    ABCValidator,
    AudioUploader,
    AuthError,
    BaseContext,
    BaseUploader,
    BotTypes,
    CallableValidator,
    Callback,
    CtxStorage,
    DelayedTask,
    DocMessagesUploader,
    DocUploader,
    DocWallUploader,
    EqualsValidator,
    GraffitiUploader,
    IsInstanceValidator,
    Keyboard,
    KeyboardButtonColor,
    Location,
    LoopWrapper,
    OpenAppEvent,
    OpenLink,
    OpenLinkEvent,
    PhotoChatFaviconUploader,
    PhotoFaviconUploader,
    PhotoMarketUploader,
    PhotoMessageUploader,
    PhotoToAlbumUploader,
    PhotoUploader,
    PhotoWallUploader,
    ShowSnackbarEvent,
    TemplateElement,
    Text,
    UserAuth,
    UserTypes,
    VideoUploader,
    VKApps,
    VKPay,
    VoiceMessageUploader,
    keyboard_gen,
    load_blueprints_from_package,
    run_in_task,
    run_sync,
    template_gen,
    vkscript,
)

event_types = GroupTypes

__all__ = (
    "ABCAction",
    "ABCAPI",
    "ABCBlueprint",
    "ABCDispenseView",
    "ABCErrorHandler",
    "ABCFramework",
    "ABCHandler",
    "ABCHTTPClient",
    "ABCPolling",
    "ABCRequestRescheduler",
    "ABCRequestValidator",
    "ABCResponseValidator",
    "ABCRouter",
    "ABCRule",
    "ABCStateDispenser",
    "ABCStorage",
    "ABCTokenGenerator",
    "ABCValidator",
    "ABCView",
    "AiohttpClient",
    "AndRule",
    "API",
    "AudioUploader",
    "AuthError",
    "BaseContext",
    "BaseMiddleware",
    "BaseReturnManager",
    "BaseStateGroup",
    "BaseUploader",
    "BlockingRequestRescheduler",
    "Bot",
    "BotBlueprint",
    "BotPolling",
    "BotTypes",
    "BuiltinStateDispenser",
    "CallableValidator",
    "Callback",
    "CaptchaError",
    "CodeException",
    "ConsistentTokenGenerator",
    "CtxStorage",
    "DEFAULT_REQUEST_VALIDATORS",
    "DEFAULT_RESPONSE_VALIDATORS",
    "DelayedTask",
    "DocMessagesUploader",
    "DocUploader",
    "DocWallUploader",
    "EMPTY_KEYBOARD",
    "EqualsValidator",
    "ErrorHandler",
    "get_token_generator",
    "GraffitiUploader",
    "GroupEventType",
    "IsInstanceValidator",
    "keyboard_gen",
    "Keyboard",
    "KeyboardButtonColor",
    "load_blueprints_from_package",
    "Location",
    "LoopWrapper",
    "MiddlewareError",
    "NotRule",
    "OpenAppEvent",
    "OpenLink",
    "OpenLinkEvent",
    "OrRule",
    "PhotoChatFaviconUploader",
    "PhotoFaviconUploader",
    "PhotoMarketUploader",
    "PhotoMessageUploader",
    "PhotoToAlbumUploader",
    "PhotoUploader",
    "PhotoWallUploader",
    "Router",
    "run_in_task",
    "run_multibot",
    "run_sync",
    "ShowSnackbarEvent",
    "SingleAiohttpClient",
    "SingleTokenGenerator",
    "StatePeer",
    "swear",
    "template_gen",
    "TemplateElement",
    "Text",
    "Token",
    "User",
    "UserAuth",
    "UserBlueprint",
    "UserEventType",
    "UserPolling",
    "UserTypes",
    "VideoUploader",
    "VKAPIError",
    "VKApps",
    "VKPay",
    "vkscript",
    "VoiceMessageUploader",
)



from abc import ABC, abstractmethod
from typing import Any, NoReturn, Union


class ABCRequestValidator(ABC):
    """Abstract Response Validator class
    Documentation: https://github.com/vkbottle/vkbottle/blob/master/docs/low-level/api/response-validator.md
    """

    @abstractmethod
    async def validate(self, response: Any) -> Union[Any, NoReturn]:
        pass



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Optional, Type, TypeVar, Union

from vkbottle.api.abc import ABCAPI
from vkbottle.dispatch.middlewares import BaseMiddleware
from vkbottle.modules import logger

if TYPE_CHECKING:
    from vkbottle.dispatch.dispenser.abc import ABCStateDispenser
    from vkbottle.dispatch.handlers import ABCHandler
    from vkbottle.dispatch.return_manager import BaseReturnManager

    Handlers = Union[List["ABCHandler"], Dict[Any, List]]

DEFAULT_STATE_KEY = "peer_id"


T_contra = TypeVar("T_contra", list, dict, contravariant=True)


class ABCView(ABC, Generic[T_contra]):
    handlers: "Handlers"
    middlewares: List[Type["BaseMiddleware"]]
    handler_return_manager: "BaseReturnManager"

    @abstractmethod
    def __init__(self):
        self.handlers = []
        self.middlewares = []
        self.handler_return_manager = None  # type: ignore

    @abstractmethod
    async def process_event(self, event: T_contra) -> bool:
        pass

    async def pre_middleware(
        self,
        event: T_contra,
        context_variables: Optional[dict] = None,
    ) -> Optional[List[BaseMiddleware]]:
        """Run all of the pre middleware methods and return an exception if any error occurs"""
        mw_instances = []

        for middleware in self.middlewares:
            mw_instance = middleware(event, view=self)
            await mw_instance.pre()
            if not mw_instance.can_forward:
                logger.debug(f"{mw_instance} pre returned error {mw_instance.error}")
                return None

            mw_instances.append(mw_instance)

            if context_variables is not None:
                context_variables.update(mw_instance.context_update)

        return mw_instances

    async def post_middleware(
        self,
        mw_instances: List[BaseMiddleware],
        handle_responses: Optional[List] = None,
        handlers: Optional[List["ABCHandler"]] = None,
    ):
        for middleware in mw_instances:
            # Update or leave value
            middleware.handle_responses = handle_responses or middleware.handle_responses
            middleware.handlers = handlers or middleware.handlers

            await middleware.post()
            if not middleware.can_forward:
                logger.debug(f"{middleware} post returned error {middleware.error}")
                return middleware.error

    @abstractmethod
    async def handle_event(
        self,
        event: T_contra,
        ctx_api: "ABCAPI",
        state_dispenser: "ABCStateDispenser",
    ) -> None:
        pass

    def register_middleware(self, middleware: Type[BaseMiddleware]):
        try:
            if not issubclass(middleware, BaseMiddleware):
                raise ValueError("Argument is not a subclass of BaseMiddleware")
        except TypeError:
            raise ValueError("Argument is not a class")
        self.middlewares.append(middleware)

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"handlers={self.handlers} "
            f"middlewares={self.middlewares} "
            f"handler_return_manager={self.handler_return_manager}"
        )



from .message import ABCUserMessageView, UserMessageView
from .raw import RawUserEventView, UserHandlerBasement

__all__ = ("ABCUserMessageView", "RawUserEventView", "UserMessageView", "UserHandlerBasement")



from .abc import ABCRequestValidator


class TranslateFriendlyTypesRequestValidator(ABCRequestValidator):
    async def validate(self, response: dict) -> dict:
        for k, v in response.items():
            # translate python-list to vk array-like type
            if isinstance(v, list):
                response[k] = ",".join(str(e) for e in v)
        return response



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from vkbottle.http import ABCHTTPClient

    from .request_rescheduler import ABCRequestRescheduler


class ABCAPI(ABC):
    """Abstract API class
    Documentation: https://github.com/vkbottle/vkbottle/blob/master/docs/low-level/api/api.md
    """

    http_client: "ABCHTTPClient"
    ignore_errors: bool
    request_rescheduler: "ABCRequestRescheduler"

    @abstractmethod
    async def request(self, method: str, data: dict) -> dict:
        """Makes a single request opening a session"""

    @abstractmethod
    async def validate_response(self, method: str, data: dict, response: Any) -> Any:
        pass

    @abstractmethod
    async def validate_request(self, request: Any) -> Any:
        pass



from typing import Optional

from pydantic import BaseModel, Field


class ShowSnackbarEvent(BaseModel):
    type: str = Field("show_snackbar", const=True)
    text: str


class OpenLinkEvent(BaseModel):
    type: str = Field("open_link", const=True)
    link: str


class OpenAppEvent(BaseModel):
    type: str = Field("open_app", const=True)
    owner_id: Optional[int] = None
    app_id: int
    hash: str


__all__ = ("ShowSnackbarEvent", "OpenLinkEvent", "OpenAppEvent")



from .dispense import ABCDispenseView
from .message import ABCMessageView
from .raw import ABCRawEventView
from .view import ABCView

__all__ = ("ABCDispenseView", "ABCMessageView", "ABCRawEventView", "ABCView")



from abc import ABC, abstractmethod
from typing import Any, Hashable


class ABCStorage(ABC):
    """Abstract storage class
    Documentation: https://github.com/vkbottle/vkbottle/blob/master/docs/tools/storage.md
    """

    @abstractmethod
    def get(self, key: Hashable) -> Any:
        pass

    @abstractmethod
    def set(self, key: Hashable, value: Any) -> None:
        pass

    @abstractmethod
    def delete(self, key: Hashable) -> None:
        pass

    @abstractmethod
    def contains(self, key: Hashable) -> bool:
        pass

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def __contains__(self, item: str) -> bool:
        return self.contains(item)

    def __setitem__(self, key: Hashable, value: Any) -> None:
        self.set(key, value)

    def __getitem__(self, key: Hashable) -> Any:
        return self.get(key)



from .message import ABCBotMessageView, BotMessageView
from .raw import BotHandlerBasement, RawBotEventView

__all__ = ("ABCBotMessageView", "BotMessageView", "BotHandlerBasement", "RawBotEventView")



from typing import TYPE_CHECKING, List, Optional

from vkbottle_types.events.bot_events import MessageNew
from vkbottle_types.objects import ClientInfoForBots, MessagesForward

from ..base import BaseMessageMin

if TYPE_CHECKING:

    from vkbottle.api import ABCAPI


class MessageMin(BaseMessageMin):
    group_id: Optional[int] = None
    client_info: Optional["ClientInfoForBots"] = None

    async def reply(
        self,
        message: Optional[str] = None,
        attachment: Optional[str] = None,
        **kwargs,
    ) -> int:
        locals().update(kwargs)

        data = {k: v for k, v in locals().items() if k not in ("self", "kwargs") and v is not None}
        data["peer_id"] = self.peer_id
        data["forward"] = MessagesForward(
            conversation_message_ids=[self.conversation_message_id],
            peer_id=self.peer_id,
            is_reply=True,
        ).json()

        return await self.answer(**data)

    async def forward(
        self,
        message: Optional[str] = None,
        attachment: Optional[str] = None,
        forward_message_ids: Optional[List[int]] = None,
        **kwargs,
    ) -> int:
        locals().update(kwargs)

        data = {
            k: v
            for k, v in locals().items()
            if k not in ("self", "kwargs", "forward_message_ids") and v is not None
        }
        if not forward_message_ids:
            forward_message_ids = [self.conversation_message_id]
        data["forward"] = MessagesForward(
            conversation_message_ids=forward_message_ids, peer_id=self.peer_id
        ).json()

        return await self.answer(**data)


def message_min(event: dict, ctx_api: "ABCAPI") -> "MessageMin":
    update = MessageNew(**event)

    if update.object.message is None:
        raise RuntimeError("Please set longpoll to latest version")

    message = MessageMin(
        **update.object.message.dict(),
        client_info=update.object.client_info.dict(),
        group_id=update.group_id,
    )
    setattr(message, "unprepared_ctx_api", ctx_api)
    return message



from time import sleep as blocking_sleep
from typing import TYPE_CHECKING, Any, Union

from vkbottle.modules import logger

from .abc import ABCRequestRescheduler

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI, API


DEFAULT_DELAY = 5


class BlockingRequestRescheduler(ABCRequestRescheduler):
    def __init__(self, delay: int = DEFAULT_DELAY):
        self.delay = delay

    async def reschedule(
        self,
        ctx_api: Union["ABCAPI", "API"],
        method: str,
        data: dict,
        recent_response: Any,
    ) -> dict:
        logger.debug(
            "Usage of blocking rescheduler is assumed when VK doesn't respond to "
            "all requests for an amount of time. Starting..."
        )

        attempt_number = 1
        while not isinstance(recent_response, dict):
            logger.info(f"Attempt number {attempt_number}. Making request...")
            blocking_sleep(self.delay * attempt_number)
            recent_response = await ctx_api.request(method, data)
            attempt_number += 1
            logger.debug(f"Attempt succeed? - {isinstance(recent_response, dict)}")

        logger.info(f"Finally succeed after {self.delay ** attempt_number} seconds")
        return recent_response



from abc import ABC
from typing import TYPE_CHECKING, Generic, Optional, TypeVar

from vkbottle_types.events import UserEventType

from vkbottle.dispatch.return_manager.user import UserMessageReturnHandler
from vkbottle.dispatch.views.abc.message import ABCMessageView
from vkbottle.tools.dev.mini_types.user import message_min

if TYPE_CHECKING:
    from vkbottle.tools.dev.mini_types.user import MessageMin


T_contra = TypeVar("T_contra", contravariant=True)


class ABCUserMessageView(ABCMessageView[list, T_contra], ABC, Generic[T_contra]):
    def __init__(self):
        super().__init__()
        self.handler_return_manager = UserMessageReturnHandler()

    @staticmethod
    def get_event_type(event: list) -> int:
        return event[0]

    @staticmethod
    async def get_message(event: list, ctx_api) -> "MessageMin":
        return await message_min(event[1], ctx_api)

    async def process_event(self, event: list) -> bool:
        if not (self.handlers or self.middlewares):
            return False
        try:
            event_type = UserEventType(self.get_event_type(event))
        except ValueError:
            event_type = UserEventType.UNDEFINED_EVENT
        return event_type == UserEventType.MESSAGE_NEW


class UserMessageView(ABCUserMessageView["MessageMin"]):
    def get_state_key(self, message: "MessageMin") -> Optional[int]:
        return getattr(message, self.state_source_key, None)



import ast
from inspect import getsource
from typing import Callable, Dict, Type, Union


class ConverterError(Exception):
    pass


class Converter:
    """Translate Python into the VKScript with AST"""

    def __init__(self):
        self.definitions: Dict[ast.AST, Callable] = {}

    def __call__(self, for_definition: Union[Type[ast.AST], ast.AST]):
        def decorator(func):
            self.definitions[for_definition] = func
            return func

        return decorator

    def find_definition(self, d):
        if d.__class__ not in self.definitions:
            raise ConverterError(
                f"Definition for {d.__class__} is undefined. Maybe vkscript doesn't support it"
            )
        return self.definitions[d.__class__](d)

    def scriptify(self, func: Callable, **values) -> str:
        """Translate function to VKScript"""
        source = getsource(func)
        code: ast.FunctionDef = ast.parse(source).body[0]  # type: ignore
        args = [a.arg for a in code.args.args]
        args.pop(0)
        if any(v not in values for v in args):
            raise ConverterError(
                "All values should be passed to func. Predefined kwargs are not allowed"
            )
        values_assignments = [f"var {k}={v!r};" for k, v in values.items()]
        return "".join(values_assignments) + "".join(
            self.find_definition(line) for line in code.body
        )



from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Generic, List, TypeVar

if TYPE_CHECKING:
    from vkbottle.api.abc import ABCAPI
    from vkbottle.dispatch.dispenser.abc import ABCStateDispenser

from vkbottle.modules import logger

from .view import ABCView

T_contra = TypeVar("T_contra", list, dict, contravariant=True)


class ABCRawEventView(ABCView[T_contra], Generic[T_contra]):
    handlers: Dict[Any, List]

    @abstractmethod
    def get_handler_basements(self, event: T_contra):
        pass

    @abstractmethod
    def get_event_model(self, handler_basement, event: T_contra):
        pass

    @staticmethod
    @abstractmethod
    def get_event_type(event: T_contra):
        pass

    async def handle_event(
        self, event: T_contra, ctx_api: "ABCAPI", state_dispenser: "ABCStateDispenser"
    ) -> Any:
        logger.debug("Handling event ({}) with message view".format(self.get_event_type(event)))

        context_variables: dict = {}
        handle_responses = []
        handlers = []

        mw_instances = await self.pre_middleware(event, context_variables)
        if mw_instances is None:
            logger.info("Handling stopped, pre_middleware returned error")
            return

        for handler_basement in self.get_handler_basements(event):
            event_model = self.get_event_model(handler_basement, event)

            if isinstance(event_model, dict):
                event_model["ctx_api"] = ctx_api
            else:
                setattr(event_model, "unprepared_ctx_api", ctx_api)

            result = await handler_basement.handler.filter(event_model)
            logger.debug("Handler {} returned {}".format(handler_basement.handler, result))

            if result is False:
                continue

            elif isinstance(result, dict):
                context_variables.update(result)

            handler_response = await handler_basement.handler.handle(
                event_model, **context_variables
            )
            handle_responses.append(handler_response)
            handlers.append(handler_basement.handler)

            return_handler = self.handler_return_manager.get_handler(handler_response)
            if return_handler is not None:
                await return_handler(
                    self.handler_return_manager,
                    handler_response,
                    event_model,
                    context_variables,
                )

            if handler_basement.handler.blocking:
                break

        await self.post_middleware(mw_instances, handle_responses, handlers)



import asyncio
from asyncio import AbstractEventLoop, get_event_loop
from typing import TYPE_CHECKING, Any, Callable, Coroutine, List, Optional, Union

from vkbottle.modules import logger

from .auto_reload import watch_to_reload
from .delayed_task import DelayedTask

if TYPE_CHECKING:
    Task = Coroutine[Any, Any, Any]


class LoopWrapper:
    """Loop Wrapper for vkbottle manages startup, shutdown and main tasks,
    creates loop and runs it forever"""

    def __init__(
        self,
        *,
        on_startup: Optional[List["Task"]] = None,
        on_shutdown: Optional[List["Task"]] = None,
        auto_reload: Optional[bool] = None,
        auto_reload_dir: Optional[str] = None,
        tasks: Optional[List["Task"]] = None,
    ):
        self.on_startup = on_startup or []
        self.on_shutdown = on_shutdown or []
        self.auto_reload = auto_reload or False
        self.auto_reload_dir = auto_reload_dir or "."
        self.tasks = tasks or []

    def run_forever(self, loop: Optional[AbstractEventLoop] = None):
        """Runs startup tasks and makes the loop running forever"""

        if not len(self.tasks):
            logger.warning("You ran loop with 0 tasks. Is it ok?")

        loop = loop or get_event_loop()

        try:
            [loop.run_until_complete(startup_task) for startup_task in self.on_startup]

            if self.auto_reload:
                loop.create_task(watch_to_reload(self.auto_reload_dir))

            for task in self.tasks:
                loop.create_task(task)

            loop.run_forever()
        except KeyboardInterrupt:
            logger.info("Keyboard Interrupt")
        finally:
            [loop.run_until_complete(shutdown_task) for shutdown_task in self.on_shutdown]
            if loop.is_running():
                loop.close()

    def add_task(self, task: Union["Task", Callable[..., "Task"]]):
        """Adds tasks to be ran in run_forever
        :param task: coroutine / coroutine function with zero arguments
        """

        if asyncio.iscoroutinefunction(task) or isinstance(task, DelayedTask):  # type: ignore
            self.tasks.append(task())  # type: ignore
        elif asyncio.iscoroutine(task):  # type: ignore
            self.tasks.append(task)  # type: ignore
        else:
            raise TypeError("Task should be coroutine or coroutine function")

    def interval(
        self, seconds: int = 0, minutes: int = 0, hours: int = 0, days: int = 0
    ) -> Callable[[Callable], Callable]:
        """A tiny template to wrap repeated tasks with decorator
        >>> lw = LoopWrapper()
        >>> @lw.interval(seconds=5)
        >>> async def repeated_function():
        >>>     print("This will be logged every five seconds")
        >>> lw.run_forever()
        """

        seconds += minutes * 60
        seconds += hours * 60 * 60
        seconds += days * 24 * 60 * 60

        def decorator(func: Callable):
            self.add_task(DelayedTask(seconds, func))
            return func

        return decorator

    def timer(
        self, seconds: int = 0, minutes: int = 0, hours: int = 0, days: int = 0
    ) -> Callable[[Callable], Callable]:
        """A tiny template to wrap tasks with timer
        >>> lw = LoopWrapper()
        >>> @lw.timer(seconds=5)
        >>> async def delayed_function():
        >>>     print("This will after 5 seconds")
        >>> lw.run_forever()
        """
        seconds += minutes * 60
        seconds += hours * 60 * 60
        seconds += days * 24 * 60 * 60

        def decorator(func: Callable):
            self.add_task(DelayedTask(seconds, func, do_break=True))
            return func

        return decorator
